@@ -63,6 +63,14 @@ History:
                                         -swap is instance(float,int,...) to not(isinstance(AD))
   Author: w.x.chan1986@gmail.com 14Oct2019           - v3.8.0
                                         -change dOrder to kwarg
+  Author: w.x.chan1986@gmail.com 23Oct2019           - v3.9.0
+                                        -add function statistics for getting stats of calculation 
+  Author: w.x.chan1986@gmail.com 18Nov2019           - v3.9.2
+                                        -changed to logging 
+  Author: w.x.chan1986@gmail.com 18Nov2019           - v3.9.3
+                                        -set call(x,xOrder=None)
+                                        -added function setNode(self,name), and recordNonde into x
+
 
 '''
 
@@ -82,19 +90,84 @@ Flexible functions accepts user-defined function and turn them into callable obj
 
 
 '''
-_version='3.8.0'
-print('autoD version',_version)
+_version='3.9.3'
+import logging
+logger = logging.getLogger('autoD v'+_version)
+logger.info('autoD version '+_version)
 
 import numpy as np
 '''
 --------------------Main Class-----------------
 '''
+class Statistics:
+    def __new__(cls, *args, **kwargs):
+        return super().__new__(cls)
+    def __init__(self):
+        self.stats={}
+        self.specialkey={}
+    def dictReport(self,d,addstr=''):
+        returnSTR=''
+        for var in d:
+            if var=='dvariables' or var=='':
+                continue
+            if isinstance(d[var],dict):
+                returnSTR+=addstr+var+':'
+                if '' in d[var]:
+                    returnSTR+=str(d[var][''])
+                returnSTR+='\n'
+                returnSTR+=self.dictReport(d[var],addstr=addstr+'    d')
+            else:
+                returnSTR+=addstr+var+':'+str(d[var])+'\n'
+        return returnSTR
+    def __repr__(self):
+        return 'STATISTICS::\n'+self.dictReport(self.stats,addstr='    ')
+    def count(self,key,countNumber,*args):
+        if key in self.specialkey:
+            self.specialkey[key](self.stats,key,countNumber,*args)
+        elif key in self.stats:
+            self.stats[key]+=countNumber
+        else:
+            self.stats[key]=countNumber
+        return;
+    def uncount(self,key,countNumber,*args):
+        if key in self.specialkey:
+            self.specialkey[key](self.stats,key,-countNumber,*args)
+        elif key in self.stats:
+            self.stats[key]-=countNumber
+        else:
+            self.stats[key]=-countNumber
+        return;
+    def countVariable(self,key,countNumber,dOrder):
+        if key not in self.specialkey:
+            self.specialkey[key]=self.countVariable
+        if key not in self.stats:
+            self.stats[key]={'dvariables':[]}
+        dvar=self.stats[key]['dvariables']
+        dvalue=list(np.zeros(len(self.stats[key]['dvariables']),dtype=int))
+        for var in dOrder:
+            if dOrder[var]>0:
+                if var in dvar:
+                    dvalue[dvar.index(var)]=dOrder[var]
+                else:
+                    dvar.append(var)
+                    dvalue.append(dOrder[var])
+        dkey=''
+        for n in range(len(dvar)):
+            if dvalue[n]>0:
+                dkey+=dvar[n]+str(dvalue[n])
+        if dkey in self.stats[key]:
+            self.stats[key][dkey]+=countNumber
+        else:
+            self.stats[key][dkey]=countNumber
+        return;
 class AD:
     def defaultDebugSwitch(self,x,dOrder,result):
         return True
     debugPrintout=False
     debugName=''
     debugSwitchFunc=defaultDebugSwitch
+    dependent=['ALL']
+    node=False
     def __pow__(self, val):
         return Power(self,val)
     def __rpow__(self, val):
@@ -131,14 +204,36 @@ class AD:
             return 'autoD function'
         strOut=','.join(self.dependent)
         return 'autoD function('+strOut+')'
+    def setID(self):
+        try:
+            autoDid=str(self.name)
+        except:
+            self.name=repr(self)
+            autoDid=str(self.name)
+        return autoDid
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        if 'ALL' not in self.dependent:
+            for var in dOrder:
+                if dOrder[var]>0 and not(var in self.dependent):
+                    return statsDict
+        statsDict.countVariable(self.setID(),1,dOrder)
+        return statsDict
     def debugPrint(self,x,dOrder,result):
         if self.debugPrintout and self.debugSwitchFunc(x,dOrder,result):
-            print(self.debugName,'@',x)
-            print('    differential=',dOrder)
-            print('    value=',result)
+            logger.debug(str(self.debugName)+' @ '+str(x))
+            logger.debug('    differential= '+str(dOrder))
+            logger.debug('    value= '+str(result))
         return;
-    
-        
+    def setNode(self,name):
+        if name in self.dependent:
+            raise Exception(repr(name)+' is already defined.')
+        if self.node:
+            logger.warning('Redefining variable from '+repr(self.node)+' to '+repr(name))
+        self.node=name
 '''
 #---------------Basic Functions-------------------------------#
 '''
@@ -160,19 +255,46 @@ class Differentiate(AD):
             self.dependent=func.dependent[:]
         except AttributeError:
             self.dependent=['ALL']
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
         new_dOrder=dOrder.copy()
         for var in self.inputorder:
             if var in dOrder:
                 new_dOrder[var]=self.inputorder[var]+dOrder[var]
             else:
-                new_dOrder[var]=self.inputorder[var]+0
+                new_dOrder[var]=self.inputorder[var]
+        if 'ALL' not in self.dependent:
+            for var in new_dOrder:
+                if new_dOrder[var]>0 and (var not in self.dependent):
+                    return statsDict
+        self.inputFunc.statistics(statsDict,new_dOrder)
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        recordNode=False
+        if self.node and sum(dOrder.values())==0:
+            if self.node in x:
+                return x[self.node]
+            else:
+                recordNode=True
+        new_dOrder=dOrder.copy()
+        for var in self.inputorder:
+            if var in dOrder:
+                new_dOrder[var]=self.inputorder[var]+dOrder[var]
+            else:
+                new_dOrder[var]=self.inputorder[var]
         if 'ALL' not in self.dependent:
             for var in new_dOrder:
                 if new_dOrder[var]>0 and (var not in self.dependent):
                     self.debugPrint(x,dOrder,0.)
                     return 0.
         result=self.inputFunc(x,new_dOrder)
+        if recordNode:
+            x[self.node]=result
         self.debugPrint(x,dOrder,result)
         return result
         
@@ -198,8 +320,28 @@ class Addition(AD):
                         self.dependent.append(dependent)
             except AttributeError:
                 self.dependent=['ALL']
-            
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        if 'ALL' not in self.dependent:
+            for var in dOrder:
+                if dOrder[var]>0 and (var not in self.dependent):
+                    return statsDict
+        statsDict.count('+',len(self.funcList))
+        for n in range(len(self.funcList)):
+            self.funcList[n].statistics(statsDict,dOrder)
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        recordNode=False
+        if self.node and sum(dOrder.values())==0:
+            if self.node in x:
+                return x[self.node]
+            else:
+                recordNode=True
         if 'ALL' not in self.dependent:
             for var in dOrder:
                 if dOrder[var]>0 and (var not in self.dependent):
@@ -209,6 +351,8 @@ class Addition(AD):
         for n in range(len(self.funcList)):
             temp.append(self.funcList[n](x,dOrder))
         result=sum(temp)
+        if recordNode:
+            x[self.node]=result
         self.debugPrint(x,dOrder,result)
         return result
               
@@ -241,7 +385,47 @@ class Multiply(AD):
             except AttributeError:
                 self.dependent=['ALL']
         self.rdOL=rotatingdOrderList(len(self.funcList))
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        if 'ALL' not in self.dependent:
+            for var in dOrder:
+                if dOrder[var]>0 and (var not in self.dependent):
+                    return statsDict
+        dOrderList,keyList=splitdOrder(dOrder)
+        pastCalculation={}
+        if len(dOrderList)==0:
+            statsDict.count('*',len(self.funcList))
+            for n in range(len(self.funcList)):
+                self.funcList[n].statistics(statsDict)
+            return statsDict
+        self.rdOL.reset(dOrderList)
+        while not(self.rdOL.end):
+            temp_dOrderList=self.rdOL.get()
+            for n in range(len(self.funcList)):
+                pastCalculationKey=str(n)+' '+''.join(map(str,temp_dOrderList[n]))
+                if pastCalculationKey in pastCalculation:
+                    statsDict.count('*',1)
+                else:
+                    temp_dOrder=mergedOrder(temp_dOrderList[n],keyList)
+                    self.funcList[n].statistics(statsDict,temp_dOrder)
+                    pastCalculation[pastCalculationKey]=0
+                    statsDict.count('*',1)
+            statsDict.count('+',1)
+            self.rdOL.incr()
+        statsDict.uncount('+',1)
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        recordNode=False
+        if self.node and sum(dOrder.values())==0:
+            if self.node in x:
+                return x[self.node]
+            else:
+                recordNode=True
         if 'ALL' not in self.dependent:
             for var in dOrder:
                 if dOrder[var]>0 and (var not in self.dependent):
@@ -272,6 +456,8 @@ class Multiply(AD):
             addList.append(mul)
             self.rdOL.incr()
         result=sum(addList)
+        if recordNode:
+            x[self.node]=result
         self.debugPrint(x,dOrder,result)
         return result
 
@@ -302,7 +488,64 @@ class Power(AD):
         else:
             self.new_exp=None
         self.rdOL=rotatingdOrderListPower()
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        if 'ALL' not in self.dependent:
+            for var in dOrder:
+                if dOrder[var]>0 and (var not in self.dependent):
+                    return statsDict
+        if self.pow==1:
+            self.func.statistics(statsDict,dOrder)
+            return statsDict
+        elif self.pow==0:
+            Constant(1.).statistics(statsDict,dOrder)
+            return statsDict
+        elif not(self.new_exp==None):
+            self.new_exp.statistics(statsDict,dOrder)
+            return statsDict
+        dOrderList,keyList=splitdOrder(dOrder)
+        if len(dOrderList)==0:
+            statsDict.count('**',1,self.pow)
+            self.func.statistics(statsDict,dOrder)
+            return statsDict
+        self.rdOL.reset(dOrderList)
+        pastCalculation={}
+        while not(self.rdOL.end):
+            temp_dOrderList=self.rdOL.get()
+            count=0
+            for n in range(len(temp_dOrderList)):
+                if len(temp_dOrderList[n])!=0:
+                    statsDict.count('*',1)
+                    count+=1
+            if (self.pow-count)!=0:
+                statsDict.count('*',1)
+                statsDict.count('**',1,self.pow-count)
+            for n in range(len(temp_dOrderList)):
+                if len(temp_dOrderList[n])!=0:
+                    pastCalculationKey=''.join(map(str, temp_dOrderList[n]))
+                    if pastCalculationKey in pastCalculation:
+                        statsDict.count('*',1)
+                    else:
+                        temp_dOrder=mergedOrder(temp_dOrderList[n],keyList)
+                        self.func.statistics(statsDict,temp_dOrder)
+                        pastCalculation[pastCalculationKey]=0
+                        statsDict.count('*',1)
+            statsDict.count('+',1)
+            self.rdOL.incr()
+        statsDict.uncount('+',1)
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        recordNode=False
+        if self.node and sum(dOrder.values())==0:
+            if self.node in x:
+                return x[self.node]
+            else:
+                recordNode=True
         if 'ALL' not in self.dependent:
             for var in dOrder:
                 if dOrder[var]>0 and (var not in self.dependent):
@@ -352,6 +595,8 @@ class Power(AD):
                 addList.append(mul)
             self.rdOL.incr()
         result=sum(addList)
+        if recordNode:
+            x[self.node]=result
         self.debugPrint(x,dOrder,result)
         return result
     
@@ -372,7 +617,47 @@ class Exp(AD):
         except AttributeError:
             self.dependent=['ALL']
         self.rdOL=rotatingdOrderListPower()
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        if 'ALL' not in self.dependent:
+            for var in dOrder:
+                if dOrder[var]>0 and (var not in self.dependent):
+                    return statsDict
+        dOrderList,keyList=splitdOrder(dOrder)
+        statsDict.count('**',1)
+        self.func.statistics(statsDict)
+        if len(dOrderList)==0:
+            return statsDict
+        self.rdOL.reset(dOrderList)
+        pastCalculation={}
+        while not(self.rdOL.end):
+            temp_dOrderList=self.rdOL.get()
+            for n in range(len(temp_dOrderList)):
+                if len(temp_dOrderList[n])!=0:
+                    pastCalculationKey=''.join(map(str,temp_dOrderList[n]))
+                    if pastCalculationKey in pastCalculation:
+                        statsDict.count('*',1)
+                    else:
+                        temp_dOrder=mergedOrder(temp_dOrderList[n],keyList)
+                        self.func.statistics(statsDict,temp_dOrder)
+                        pastCalculation[pastCalculationKey]=0
+                        statsDict.count('*',1)
+            statsDict.count('+',1)
+            self.rdOL.incr()
+        statsDict.uncount('+',1)
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        recordNode=False
+        if self.node and sum(dOrder.values())==0:
+            if self.node in x:
+                return x[self.node]
+            else:
+                recordNode=True
         if 'ALL' not in self.dependent:
             for var in dOrder:
                 if dOrder[var]>0 and (var not in self.dependent):
@@ -402,6 +687,8 @@ class Exp(AD):
             addList.append(mul)
             self.rdOL.incr()
         result=sum(addList)*exp_value
+        if recordNode:
+            x[self.node]=result
         self.debugPrint(x,dOrder,result)
         return result
         
@@ -422,7 +709,56 @@ class Ln(AD):
         except AttributeError:
             self.dependent=['ALL']
         self.rdOL=rotatingdOrderListPower()
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        if 'ALL' not in self.dependent:
+            for var in dOrder:
+                if dOrder[var]>0 and (var not in self.dependent):
+                    return statsDict
+        dOrderList,keyList=splitdOrder(dOrder)
+        if len(dOrderList)==0:
+            statsDict.count('log',1)
+            self.func.statistics(statsDict)
+            return statsDict
+        self.rdOL.reset(dOrderList)
+        pastCalculation={}
+        while not(self.rdOL.end):
+            temp_dOrderList=self.rdOL.get()
+            count=0
+            for n in range(len(temp_dOrderList)):
+                if len(temp_dOrderList[n])!=0:
+                    if count!=0:
+                        statsDict.count('*',1)
+                    count-=1
+            statsDict.count('*',1)
+            statsDict.count('**',1,count)
+            self.func.statistics(statsDict)
+            for n in range(len(temp_dOrderList)):
+                if len(temp_dOrderList[n])!=0:
+                    pastCalculationKey=''.join(map(str,temp_dOrderList[n]))
+                    if pastCalculationKey in pastCalculation:
+                        statsDict.count('*',1)
+                    else:
+                        temp_dOrder=mergedOrder(temp_dOrderList[n],keyList)
+                        self.func.statistics(statsDict,temp_dOrder)
+                        pastCalculation[pastCalculationKey]=0
+                        statsDict.count('*',1)
+            statsDict.count('+',1)
+            self.rdOL.incr()
+        statsDict.uncount('+',1)
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        recordNode=False
+        if self.node and sum(dOrder.values())==0:
+            if self.node in x:
+                return x[self.node]
+            else:
+                recordNode=True
         if 'ALL' not in self.dependent:
             for var in dOrder:
                 if dOrder[var]>0 and (var not in self.dependent):
@@ -459,6 +795,8 @@ class Ln(AD):
             addList.append(mul)
             self.rdOL.incr()
         result=sum(addList)
+        if recordNode:
+            x[self.node]=result
         self.debugPrint(x,dOrder,result)
         return result
         
@@ -488,13 +826,35 @@ class Log(AD):
         else:
             self.new_ln=Ln(self.func)
             self.coef=-1./np.log(self.base)
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        if 'ALL' not in self.dependent:
+            for var in dOrder:
+                if dOrder[var]>0 and (var not in self.dependent):
+                    return statsDict
+        statsDict.count('*',1)
+        self.new_ln.statistics(statsDict,dOrder)
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        recordNode=False
+        if self.node and sum(dOrder.values())==0:
+            if self.node in x:
+                return x[self.node]
+            else:
+                recordNode=True
         if 'ALL' not in self.dependent:
             for var in dOrder:
                 if dOrder[var]>0 and (var not in self.dependent):
                     self.debugPrint(x,dOrder,0.)
                     return 0.
         result=self.coef*self.new_ln(x,dOrder)
+        if recordNode:
+            x[self.node]=result
         self.debugPrint(x,dOrder,result)
         return result
             
@@ -515,21 +875,69 @@ class Cos(AD):
         except AttributeError:
             self.dependent=['ALL']
         self.rdOL=rotatingdOrderListPower()
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        if 'ALL' not in self.dependent:
+            for var in dOrder:
+                if dOrder[var]>0 and (var not in self.dependent):
+                    return statsDict
+        dOrderList,keyList=splitdOrder(dOrder)
+        statsDict.count('cos',1)
+        self.func.statistics(statsDict)
+        if len(dOrderList)==0:
+            return statsDict
+        self.rdOL.reset(dOrderList)
+        pastCalculation={}
+        statsDict.count('sin',1)
+        while not(self.rdOL.end):
+            temp_dOrderList=self.rdOL.get()
+            count=0
+            for n in range(len(temp_dOrderList)):
+                if len(temp_dOrderList[n])!=0:
+                    count+=1
+            for n in range(len(temp_dOrderList)):
+                if len(temp_dOrderList[n])!=0:
+                    pastCalculationKey=''.join(map(str,temp_dOrderList[n]))
+                    if pastCalculationKey in pastCalculation:
+                        statsDict.count('*',1)
+                    else:
+                        temp_dOrder=mergedOrder(temp_dOrderList[n],keyList)
+                        self.func.statistics(statsDict,temp_dOrder)
+                        pastCalculation[pastCalculationKey]=0
+                        statsDict.count('*',1)
+            statsDict.count('*',1)
+            statsDict.count('/',1)
+            statsDict.count('+',1)
+            self.rdOL.incr()
+        statsDict.uncount('+',1)
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        recordNode=False
+        if self.node and sum(dOrder.values())==0:
+            if self.node in x:
+                return x[self.node]
+            else:
+                recordNode=True
         if 'ALL' not in self.dependent:
             for var in dOrder:
                 if dOrder[var]>0 and (var not in self.dependent):
                     self.debugPrint(x,dOrder,0.)
                     return 0.
         dOrderList,keyList=splitdOrder(dOrder)
-        cosValue=np.cos(self.func(x,{}))
+        funcValue=self.func(x,{})
+        cosValue=np.cos(funcValue)
         if len(dOrderList)==0:
             self.debugPrint(x,dOrder,cosValue)
             return cosValue
         self.rdOL.reset(dOrderList)
         addList=[]
         pastCalculation={}
-        sinValue=np.sin(self.func(x,{}))
+        sinValue=np.sin(funcValue)
         while not(self.rdOL.end):
             mul=1.
             temp_dOrderList=self.rdOL.get()
@@ -559,6 +967,8 @@ class Cos(AD):
             addList.append(mul)
             self.rdOL.incr()
         result=sum(addList)
+        if recordNode:
+            x[self.node]=result
         self.debugPrint(x,dOrder,result)
         return result
 class Cosh(AD):
@@ -578,8 +988,25 @@ class Cosh(AD):
         except AttributeError:
             self.dependent=['ALL']
         self.cosh=Cos(func*1j)
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        self.cosh.statistics(statsDict,dOrder)
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        recordNode=False
+        if self.node and sum(dOrder.values())==0:
+            if self.node in x:
+                return x[self.node]
+            else:
+                recordNode=True
         result=self.cosh(x,dOrder)
+        if recordNode:
+            x[self.node]=result
         self.debugPrint(x,dOrder,result)
         return result
 class Sin(AD):
@@ -599,21 +1026,69 @@ class Sin(AD):
         except AttributeError:
             self.dependent=['ALL']
         self.rdOL=rotatingdOrderListPower()
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        if 'ALL' not in self.dependent:
+            for var in dOrder:
+                if dOrder[var]>0 and (var not in self.dependent):
+                    return statsDict
+        dOrderList,keyList=splitdOrder(dOrder)
+        statsDict.count('sin',1)
+        self.func.statistics(statsDict)
+        if len(dOrderList)==0:
+            return statsDict
+        self.rdOL.reset(dOrderList)
+        pastCalculation={}
+        statsDict.count('cos',1)
+        while not(self.rdOL.end):
+            temp_dOrderList=self.rdOL.get()
+            count=0
+            for n in range(len(temp_dOrderList)):
+                if len(temp_dOrderList[n])!=0:
+                    count+=1
+            for n in range(len(temp_dOrderList)):
+                if len(temp_dOrderList[n])!=0:
+                    pastCalculationKey=''.join(map(str,temp_dOrderList[n]))
+                    if pastCalculationKey in pastCalculation:
+                        statsDict.count('*',1)
+                    else:
+                        temp_dOrder=mergedOrder(temp_dOrderList[n],keyList)
+                        self.func.statistics(statsDict,temp_dOrder)
+                        pastCalculation[pastCalculationKey]=temp_value
+                        statsDict.count('*',1)
+            statsDict.count('*',1)
+            statsDict.count('/',1)
+            statsDict.count('+',1)
+            self.rdOL.incr()
+        statsDict.uncount('+',1)
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        recordNode=False
+        if self.node and sum(dOrder.values())==0:
+            if self.node in x:
+                return x[self.node]
+            else:
+                recordNode=True
         if 'ALL' not in self.dependent:
             for var in dOrder:
                 if dOrder[var]>0 and (var not in self.dependent):
                     self.debugPrint(x,dOrder,0.)
                     return 0.
         dOrderList,keyList=splitdOrder(dOrder)
-        sinValue=np.sin(self.func(x,{}))
+        funcValue=self.func(x,{})
+        sinValue=np.sin(funcValue)
         if len(dOrderList)==0:
             self.debugPrint(x,dOrder,sinValue)
             return sinValue
         self.rdOL.reset(dOrderList)
         addList=[]
         pastCalculation={}
-        cosValue=np.cos(self.func(x,{}))
+        cosValue=np.cos(funcValue)
         while not(self.rdOL.end):
             mul=1.
             temp_dOrderList=self.rdOL.get()
@@ -643,6 +1118,8 @@ class Sin(AD):
             addList.append(mul)
             self.rdOL.incr()
         result=sum(addList)
+        if recordNode:
+            x[self.node]=result
         self.debugPrint(x,dOrder,result)
         return result
 class Sinh(AD):
@@ -662,8 +1139,25 @@ class Sinh(AD):
         except AttributeError:
             self.dependent=['ALL']
         self.sinh=-1j*Sin(func*1j)
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        self.sinh.statistics(statsDict,dOrder)
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        recordNode=False
+        if self.node and sum(dOrder.values())==0:
+            if self.node in x:
+                return x[self.node]
+            else:
+                recordNode=True
         result=self.sinh(x,dOrder)
+        if recordNode:
+            x[self.node]=result
         self.debugPrint(x,dOrder,result)
         return result
 class Tan(AD):
@@ -683,8 +1177,25 @@ class Tan(AD):
         except AttributeError:
             self.dependent=['ALL']
         self.tan=Sin(func)/Cos(func)
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        self.tan.statistics(statsDict,dOrder)
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        recordNode=False
+        if self.node and sum(dOrder.values())==0:
+            if self.node in x:
+                return x[self.node]
+            else:
+                recordNode=True
         result=self.tan(x,dOrder)
+        if recordNode:
+            x[self.node]=result
         self.debugPrint(x,dOrder,result)
         return result
 class Tanh(AD):
@@ -705,10 +1216,28 @@ class Tanh(AD):
             self.dependent=['ALL']
         self.tanh_negative=(1-Exp(-2.*func))/(1+Exp(-2.*func))
         self.tanh_positive=(Exp(2.*func)-1)/(Exp(2.*func)+1)
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        self.tanh_negative.statistics(statsDict,dOrder)
+        self.tanh_positive.statistics(statsDict,dOrder)
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        recordNode=False
+        if self.node and sum(dOrder.values())==0:
+            if self.node in x:
+                return x[self.node]
+            else:
+                recordNode=True
         result=self.tanh_negative(x,dOrder)
         if not(float('-inf')<np.abs(result)<float('inf')):
             result=self.tanh_positive(x,dOrder)
+        if recordNode:
+            x[self.node]=result
         self.debugPrint(x,dOrder,result)
         return result
 '''
@@ -730,8 +1259,25 @@ class Conjugate(AD):
             self.dependent=func.dependent[:]
         except AttributeError:
             self.dependent=['ALL']
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        statsDict.count('+',1)
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        recordNode=False
+        if self.node and sum(dOrder.values())==0:
+            if self.node in x:
+                return x[self.node]
+            else:
+                recordNode=True
         result=np.conjugate(self.func(x,dOrder))
+        if recordNode:
+            x[self.node]=result
         self.debugPrint(x,dOrder,result)
         return result
 class Real(AD):
@@ -750,8 +1296,25 @@ class Real(AD):
             self.dependent=func.dependent[:]
         except AttributeError:
             self.dependent=['ALL']
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        self.func.statistics(statsDict,dOrder)
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        recordNode=False
+        if self.node and sum(dOrder.values())==0:
+            if self.node in x:
+                return x[self.node]
+            else:
+                recordNode=True
         result=self.func(x,dOrder).real
+        if recordNode:
+            x[self.node]=result
         self.debugPrint(x,dOrder,result)
         return result
 class Imaginary(AD):
@@ -770,8 +1333,25 @@ class Imaginary(AD):
             self.dependent=func.dependent[:]
         except AttributeError:
             self.dependent=['ALL']
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        self.func.statistics(statsDict,dOrder)
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        recordNode=False
+        if self.node and sum(dOrder.values())==0:
+            if self.node in x:
+                return x[self.node]
+            else:
+                recordNode=True
         result=self.func(x,dOrder).imag
+        if recordNode:
+            x[self.node]=result
         self.debugPrint(x,dOrder,result)
         return result
 class Absolute(AD):
@@ -791,8 +1371,25 @@ class Absolute(AD):
             self.dependent=func.dependent[:]
         except AttributeError:
             self.dependent=['ALL']
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        statsDict.count('+',1)
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        recordNode=False
+        if self.node and sum(dOrder.values())==0:
+            if self.node in x:
+                return x[self.node]
+            else:
+                recordNode=True
         result=self.abs(x,dOrder)
+        if recordNode:
+            x[self.node]=result
         self.debugPrint(x,dOrder,result)
         return result
 '''
@@ -803,7 +1400,15 @@ class Constant(AD):
     def __init__(self,const):
         self.const=const
         self.dependent=[]
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
         for var in dOrder:
             if dOrder[var]>0:
                 self.debugPrint(x,dOrder,0.)
@@ -817,7 +1422,16 @@ class Scalar(AD):
     def __init__(self,name):
         self.name=name
         self.dependent=[name]
-    def __call__(self,x,dOrder={}):
+    def statistics(self,statsDict=None,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        if type(statsDict)==type(None):
+            statsDict=Statistics()
+        statsDict.count(self.name,1,{})
+        return statsDict
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
         returnX=True
         for var in dOrder:
             if dOrder[var]>0:
@@ -844,9 +1458,24 @@ class Function(AD):
         self.func=func
         self.args=args
         self.dependent=dependent
-    def __call__(self,x,dOrder={}):
+    def __call__(self,x,dOrder=None):
+        if dOrder is None:
+            dOrder={}
+        recordNode=False
+        if self.node and sum(dOrder.values())==0:
+            if self.node in x:
+                return x[self.node]
+            else:
+                recordNode=True
+        if 'ALL' not in self.dependent:
+            for var in new_dOrder:
+                if new_dOrder[var]>0 and (var not in self.dependent):
+                    self.debugPrint(x,dOrder,0.)
+                    return 0.
         args=self.args
         result=self.func(x,dOrder,*args)
+        if recordNode:
+            x[self.node]=result
         self.debugPrint(x,dOrder,result)
         return result
     def changeArgs(self,*new_args):
